# -*- coding: utf-8 -*-
"""Test the tcex.utils package."""

import datetime
import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import utility


def test_date_to_datetime():
    starting_date = datetime.datetime.now()
    tcex_instance = utility.init_tcex(clean_data=False)
    new_date = tcex_instance.utils.date_to_datetime(str(starting_date))
    assert new_date.year == starting_date.year
    assert new_date.month == starting_date.month
    assert new_date.day == starting_date.day
    assert new_date.hour == starting_date.hour

    # try the same thing again with a different timezone
    tcex_instance = utility.init_tcex(clean_data=False)
    new_date = tcex_instance.utils.date_to_datetime(str(datetime.datetime.now()), tz='UTC')
    # make sure the hour is different b/c of the different timezone
    assert new_date.hour != starting_date.hour


def test_timedelta():
    from tcex import tcex_utils
    utils = tcex_utils.TcExUtils()
    delta = utils.timedelta('2018-01-13', '2018-02-14')
    assert delta['days'] == -1
    assert delta['total_days'] == -32
    assert delta['hours'] == 0
    assert delta['total_weeks'] == -4



# -*- coding: utf-8 -*-
"""Test the TcEx app.."""

import json
import os
import sys

import pytest

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))
import utility

TEST_DATASTORE_PATH = 'testing'


def test_creation():
    tcex = utility.init_tcex(requires_tc_token=True)
    resource = tcex.resources.DataStore(tcex)
    body = {'one': 1}
    results = resource.create('organization', TEST_DATASTORE_PATH, 1, json.dumps(body)).get('data')
    assert results
    assert results['_shards']['successful'] == 1
    assert results['_type'] == TEST_DATASTORE_PATH


def test_retrieval():
    tcex = utility.init_tcex(requires_tc_token=True)
    resource = tcex.resources.DataStore(tcex)
    results = resource.read('organization', TEST_DATASTORE_PATH, 1).get('data')
    assert results
    assert results['_source'] == {'one': 1}
    assert results['found'] == True
    assert results['_type'] == TEST_DATASTORE_PATH


def test_creation_with_string_path():
    tcex = utility.init_tcex(requires_tc_token=True)
    resource = tcex.resources.DataStore(tcex)
    body = {'two': 2}
    results = resource.create('organization', TEST_DATASTORE_PATH, 'testing', json.dumps(body)).get('data')
    assert results
    assert results['_shards']['successful'] == 1
    assert results['_type'] == TEST_DATASTORE_PATH


def test_retrieval_with_string_path():
    tcex = utility.init_tcex(requires_tc_token=True)
    resource = tcex.resources.DataStore(tcex)
    results = resource.read('organization', TEST_DATASTORE_PATH, 'testing').get('data')
    assert results
    assert results['_source'] == {'two': 2}
    assert results['found'] == True
    assert results['_type'] == TEST_DATASTORE_PATH


def test_creation_path_with_spaces():
    """Try creating an entry in the datastore where the path has a space in it."""
    tcex = utility.init_tcex(requires_tc_token=True)
    resource = tcex.resources.DataStore(tcex)
    body = {'three': 3}
    results = resource.create('organization', TEST_DATASTORE_PATH, 'test ing', json.dumps(body)).get('data')
    assert results
    assert results['_shards']['successful'] == 1
    assert results['_type'] == TEST_DATASTORE_PATH


def test_retrieval_path_with_spaces():
    """Try retrieving an entry in the datastore where the path has a space in it."""
    tcex = utility.init_tcex(requires_tc_token=True)
    resource = tcex.resources.DataStore(tcex)
    results = resource.read('organization', TEST_DATASTORE_PATH, 'test ing').get('data')
    assert results
    assert results['_source'] == {'three': 3}
    assert results['found'] == True
    assert results['_type'] == TEST_DATASTORE_PATH


def test_creation_path_with_unicode():
    tcex = utility.init_tcex(requires_tc_token=True)
    resource = tcex.resources.DataStore(tcex)
    body = {'four': 4}
    results = resource.create('organization', TEST_DATASTORE_PATH, 'niño', json.dumps(body)).get('data')
    assert results
    assert results['_shards']['successful'] == 1
    assert results['_type'] == TEST_DATASTORE_PATH


def test_retrieval_path_with_unicode():
    tcex = utility.init_tcex(requires_tc_token=True)
    resource = tcex.resources.DataStore(tcex)
    results = resource.read('organization', TEST_DATASTORE_PATH, 'niño').get('data')
    assert results
    assert results['_source'] == {'four': 4}
    assert results['found'] == True
    assert results['_type'] == TEST_DATASTORE_PATH


@pytest.fixture(scope="session", autouse=True)
def cleanup(request):
    """Cleanup the datastore."""
    def clear_datastore():
        tcex = utility.init_tcex(requires_tc_token=True)
        resource = tcex.resources.DataStore(tcex)
        results = resource.delete('organization', TEST_DATASTORE_PATH, 1).get('data')
        assert results['_shards']['successful'] == 1
        assert results['result'] == 'deleted'
        results = resource.delete('organization', TEST_DATASTORE_PATH, 'testing').get('data')
        assert results['_shards']['successful'] == 1
        assert results['result'] == 'deleted'
        results = resource.delete('organization', TEST_DATASTORE_PATH, 'test ing').get('data')
        assert results['_shards']['successful'] == 1
        assert results['result'] == 'deleted'
        results = resource.delete('organization', TEST_DATASTORE_PATH, 'niño').get('data')
        assert results['_shards']['successful'] == 1
        assert results['result'] == 'deleted'
    request.addfinalizer(clear_datastore)



# -*- coding: utf-8 -*-
""" TcEx Framework Job Module """
import json
import re
import time


class TcExJob(object):
    """Job processing functionality.

    .. warning:: This module will be deprecated in version 0.9.0.

    Supports batch indicator adds and allows structured group adds.
    """

    def __init__(self, tcex):
        """Initialize the Class properties.

        Args:
            tcex (instance): An instance of TcEx class.
        """
        self._tcex = tcex

        # defaults batch settings
        self._batch_action = self._tcex.default_args.batch_action
        self._batch_chunk = self._tcex.default_args.batch_chunk
        self._batch_halt_on_error = self._tcex.default_args.batch_halt_on_error
        self._batch_poll_interval_max = self._tcex.default_args.batch_poll_interval_max
        self._batch_poll_interval = self._tcex.default_args.batch_poll_interval
        self._batch_write_type = self._tcex.default_args.batch_write_type

        # containers
        self._associations = []
        self._file_occurrences = []
        self._group_cache = {}
        self._group_cache_id = {}
        self._group_associations = []
        self._group_results = {
            'cached': [],
            'failed': [],
            'not_saved': [],
            'saved': [],
            'submitted': []
        }
        self._groups = []
        self._groups_response = []
        self._indicator_results = {
            'failed': [],
            'not_saved': [],
            'saved': [],
            'submitted': []
        }
        self._indicators = []
        self._indicators_response = []

        # batch "bad" errors
        self._batch_failures = [
            'Encountered an unexpected Exception while processing batch job',
            'would exceed the number of allowed indicators'
        ]

        # keep the batch_id
        self._indicator_batch_ids = []

    def _chunk_indicators(self):
        """Split indicator list into smaller more manageable numbers."""
        try:
            for i in xrange(0, len(self._indicators), self._batch_chunk):
                yield self._indicators[i:i + self._batch_chunk]
        except NameError:  # Python 3 uses range() instead of xrange()
            for i in range(0, len(self._indicators), self._batch_chunk):
                yield self._indicators[i:i + self._batch_chunk]

    def _group_add(self, resource_type, resource_name, owner, data=None):
        """Add a group to ThreatConnect.

        Args:
            resource_type (string): String with resource [group] type.
            resource_name (string): The name of the group.
            owner (string): The name of the TC owner for the group add.
            data (Optional [dictionary]): Any optional group parameters, tags,
                or attributes.

        Returns:
            (integer): The ID of the created resource.
        """
        resource_id = None
        if data is None:
            data = {}

        self._tcex.log.debug(u'Adding {} "{}" in owner {}'.format(
            resource_type, resource_name, owner))
        resource_body = {}
        resource = self._tcex.resource(resource_type)
        resource.http_method = 'POST'
        resource.owner = owner

        # dynamically handle additional parameters
        for param, value in data.items():
            if param in ['attribute', 'fileData', 'tag', 'type']:
                # attributes, fileData and tags are handled separately
                continue
            resource_body[param] = value
            self._tcex.log.debug(u'Adding parameter {}:{}'.format(
                param, value))

        resource.body = json.dumps(resource_body)
        results = resource.request()

        if results.get('status') == 'Success':
            resource_id = results['data']['id']  # get created group ID
            # update cache
            self._group_cache[owner][resource_type][resource_name] = resource_id

            self._group_results['saved'].append(resource_name)
            self._group_results['not_saved'].remove(resource_name)
            resource = self._tcex.resource(resource_type)
            resource.http_method = 'POST'

            # add attributes
            for attribute in data.get('attribute', []):
                self._tcex.log.debug(u'Adding attribute type "{}" to Group ID {}'.format(
                    attribute.get('type'), resource_id))
                resource.resource_id(resource_id)
                attribute_resource = resource.attributes()
                attribute_resource.body = json.dumps(attribute)
                a_results = attribute_resource.request()
                if a_results.get('status') != 'Success':
                    err = u'Failed adding attribute type "{}" with value "{}" to group "{}". ({})'
                    err = err.format(
                        attribute.get('type'), attribute.get('value'), resource_name,
                        a_results.get('response').text)
                    self._tcex.log.error(err)
                    self._tcex.exit_code = 3

            # add tags
            for tag in data.get('tag', []):
                self._tcex.log.debug(u'Adding tag ({})'.format(tag))
                resource.resource_id(resource_id)
                tag_resource = resource.tags(tag.get('name'))
                t_results = tag_resource.request()
                if t_results.get('status') != 'Success':
                    err = u'Failed adding tag "{}" ({})'.format(
                        tag, t_results.get('response').text)
                    self._tcex.log.error(err)
                    self._tcex.exit_code = 3

            # document upload
            if resource_type == 'Document' and data.get('fileData') is not None:
                try:
                    resource.upload(resource_id, data.get('fileData').encode('utf-8'))
                except UnicodeDecodeError:
                    resource.upload(resource_id, data.get('fileData'))  # py2 issue
                except AttributeError:
                    resource.upload(resource_id, data.get('fileData'))  # py3
                resource.http_method = 'POST'
                u_results = resource.request()
                if u_results['response'].status_code == 401:
                    # PUT if document already exists
                    resource = self._tcex.resource(resource_type)
                    resource.upload(resource_id, data.get('fileData'))
                    resource.http_method = 'PUT'
                    u_results = resource.request()

                if u_results.get('status') != 'Success':
                    err = u'Failed uploading document {} to group {}. ({})'.format(
                        data.get('fileName'), resource_name, u_results.get('response').text)
                    self._tcex.log.error(err)
                    self._tcex.exit_code = 3
        else:
            self._group_results['failed'].append(resource_name)
            err = u'Failed adding group "{}" ({})'.format(
                resource_name, results.get('response').text)
            self._tcex.log.error(err)
            self._tcex.exit_code = 3

        return resource_id

    def _group_delete_by_id(self, resource_type, resource_id):
        """Delete a group from ThreatConnect.

        Args:
            resource_type (string): String with resource [group] type.
            resource_id (string): The ID of the group.
        """
        # self._tcex.log.debug(u'Deleting {} with ID {}.'.format(resource_type, resource_id))
        resource = self._tcex.resource(resource_type)
        resource.http_method = 'DELETE'
        resource.resource_id(resource_id)
        results = resource.request()
        if results.get('status') != 'Success':
            self._tcex.log.warning(u'Deleted {} with ID {} failed ({}).'.format(
                resource_type, resource_id, results.get('response').text))

    def _group_delete_by_name(self, owner, resource_type, resource_name):
        """Delete group(s) from ThreatConnect by Name.

        Args:
            resource_type (string): String with resource [group] type.
            resource_name (string): The name of the group.
        """
        resource = self._tcex.resource(resource_type)
        resource.add_filter('name', '=', resource_name)
        resource.owner = owner
        for results in resource:
            if results.get('status') != 'Success':
                self._tcex.log.warning(u'Retrieve {} with name {} failed ({}).'.format(
                    resource_type, resource_name, results.get('response').text))
            else:
                # remove group from cache
                self.group_cache_remove(owner, resource_type, resource_name)
                for result in results['data']:
                    self._group_delete_by_id(resource_type, result.get('id'))

    def _indicators_replace_stub(self, owner):
        """Process Indicator data replacing associatedGroup stub values with the Group ID.

        Format of a stub ${<group name>:<group type>}.

        Args:
            indicators (list): Batch Indicator Data.
            owner (string):  The owner name for the indicator to be written.
        """
        for indicator_data in self._indicators:
            for group_value in list(indicator_data.get('associatedGroup', [])):
                if not str(group_value).startswith('$'):
                    continue

                # remove stub value from list
                indicator_data['associatedGroup'].remove(group_value)
                # extract group name and type from stub
                group_pattern = re.compile(r'^\$\((.+):(.+)\)$')
                group_data = re.search(group_pattern, group_value)
                group_name = group_data.group(1)
                group_type = group_data.group(2)
                group_id = self.group_id(group_name, owner, group_type)

                if group_id is not None:
                    self._tcex.log.debug('Replacing stub {} with id {}.'.format(
                        group_value, group_id))
                    # add group id returned from stub values
                    indicator_data['associatedGroup'].append(group_id)

    def _process_file_occurrences(self, owner):
        """Process file occurrences and write to TC API.

        Args:
            owner (string):  The owner name for the indicator to be written.
        """
        # POST /v2/indicators/files/BE7DE2F0CF48294400C714C9E28ECD01/fileOccurrences
        # supported format -> 2014-11-03T00:00:00-05:00

        resource = self._tcex.resource('File')
        resource.http_method = 'POST'
        resource.owner = owner
        resource.url = self._tcex.default_args.tc_api_path

        for occurrence in self._file_occurrences:
            # remove the hash from the dictionary and add to URI
            if occurrence.get('hash') is None:
                err = u'A hash value must be provided.'
                self._tcex.log.error(err)
                return
            resource.occurrence(occurrence.pop('hash'))
            resource.body = json.dumps(occurrence)
            resource.request()

    def _process_association(self, owner):
        """Process groups and write to TC API.

        Args:
            owner (string):  The owner name for the indicator to be written.
        """
        for a in self._associations:
            self._tcex.log.info(u'creating association: resource {} association {}'.format(
                a.get('resource_value'), a.get('association_value')))

            association_type = a.get('association_type')
            association_id = a.get('association_value')
            resource_type = a.get('resource_type')
            resource_id = a.get('resource_value')
            custom_association_name = a.get('custom_association_name')
            # get group id from group name (duplicate group names not supported)
            if resource_type in self._tcex.group_types:
                resource_id = self.group_id(resource_id, owner, resource_type)
            if association_type in self._tcex.group_types:
                association_id = self.group_id(association_id, owner, association_type)

            resource = self._tcex.resource(resource_type)
            resource.http_method = 'POST'
            resource.resource_id(resource_id)
            resource.owner = owner
            resource.url = self._tcex.default_args.tc_api_path

            ar = self._tcex.resource(association_type)
            ar.resource_id(association_id)
            if custom_association_name:
                association_resource = resource.association_custom(custom_association_name, ar)
            else:
                association_resource = resource.association_pivot(ar)
            association_results = association_resource.request()
            if association_results.get('status') != 'Success':
                err = 'Failed adding association ({}).'.format(
                    association_results.get('response').text)
                self._tcex.log.error(err)
                self._tcex.exit_code = 3

    def _process_group_association(self, owner):
        """Process group associations and write to ThreatConnect API.

        Args:
            owner (string): The owner name for the associations to be written.
        """
        for ga in self._group_associations:
            self._tcex.log.info(u'creating association: group "{}" indicator {}'.format(
                ga.get('group_name'), ga.get('indicator')))

            group_id = self.group_id(ga.get('group_name'), owner, ga.get('group_type'))
            if group_id is None:
                continue

            resource = self._tcex.resource(ga.get('indicator_type'))
            resource.http_method = 'POST'
            resource.indicator(ga.get('indicator'))
            resource.owner = owner
            resource.url = self._tcex.default_args.tc_api_path

            ar = self._tcex.resource(ga.get('group_type'))
            ar.resource_id(group_id)
            association_resource = resource.association_pivot(ar)
            association_resource.request()

    def _process_groups(self, owner, group_action):
        """Process groups and write to ThreatConnect API.

        Args:
            owner (string): The owner name for the groups to be written.
        """
        self._group_results['not_saved'] = list(self._group_results.get('submitted', []))
        for group in self._groups:
            # get group id from cache if exists
            group_name = group.get('name')
            group_type = group.get('type')
            group_id = self.group_id(group_name, owner, group_type)

            self._tcex.log.debug(u'Processing group ({})'.format(group_name))
            if group_action == 'duplicate':
                group_id = self._group_add(group_type, group_name, owner, group)
                if group_id is None and self._batch_halt_on_error:
                    self._tcex.log.info(u'Halt on error is enabled.')
                    self._tcex.exit_code = 1
                    break
                self._tcex.log.info(u'Created Group "{}" with id: {}'.format(group_name, group_id))
            elif group_action == 'replace':
                if group_id is not None:
                    # delete the group from TC and group cache
                    self._group_delete_by_name(owner, group_type, group_name)
                    self._tcex.log.info(u'Deleting Group "{}" with name: "{}"'.format(
                        group_type, group_name))

                group_id = self._group_add(group_type, group_name, owner, group)
                if group_id is None and self._batch_halt_on_error:
                    self._tcex.log.info(u'Halt on error is enabled.')
                    self._tcex.exit_code = 1
                    break
                else:
                    self.group_cache_add(group_name, owner, group_type, group_id)
                self._tcex.log.info(u'Created Group "{}" with id: {}'.format(group_name, group_id))
            elif group_action == 'skip':
                if group_id is None:
                    group_id = self._group_add(group_type, group_name, owner, group)
                    if group_id is None and self._batch_halt_on_error:
                        self._tcex.log.info(u'Halt on error is enabled.')
                        self._tcex.exit_code = 1
                        break
                    else:
                        self.group_cache_add(group_name, owner, group_type, group_id)
                    self._tcex.log.info(u'Created Group "{}" with id: {}'.format(
                        group_name, group_id))
                else:
                    # update group results tracking
                    self._group_results['cached'].append(group_name)
                    self._group_results['not_saved'].remove(group_name)
                    self._tcex.log.info(u'Skipping existing Group "{}")'.format(group_name))
            else:
                self._tcex.log.info(u'Invalid group action ({})'.format(group_action))

    def _process_indicators(self, owner, batch):
        """Process batch indicators and write to ThreatConnect API.

        Args:
            owner (string): The owner name for the indicator to be written.
            batch (boolean): Use the batch API to create indicators.
        """
        self._tcex.log.info(u'Processing {} indicators'.format(len(self._indicators)))
        if batch:
            self._process_indicators_batch(owner)
        else:
            self._process_indicators_v2(owner)

    def _process_indicators_batch(self, owner):
        """Process batch indicators and write to ThreatConnect API.

        .. Note:: Failed attributes and/or tags will not cause a batch import to fail.

        Args:
            owner (string):  The owner name for the indicator to be written.
        """
        batch_job_body = {
            'action': self._batch_action,
            'attributeWriteType': self._batch_write_type,
            'haltOnError': self._batch_halt_on_error,
            'owner': owner
        }

        halt = False
        for chunk in self._chunk_indicators():
            self._tcex.log.info(u'Batch Chunk Size: {}'.format(len(chunk)))

            # new batch resource for each chunk
            resource = self._tcex.resource('Batch')
            resource.http_method = 'POST'
            resource.url = self._tcex.default_args.tc_api_path
            resource.content_type = 'application/json'
            resource.body = json.dumps(batch_job_body)
            results = resource.request()

            if results['status'] == 'Success':
                batch_id = results.get('data')
                self._indicator_batch_ids.append(batch_id)

                resource.content_type = 'application/octet-stream'
                resource.batch_id(batch_id)
                resource.body = json.dumps(chunk)
                # results = resource.request()
                data_results = resource.request()

                if data_results['status'] == 'Success':
                    # bcs - add a small delay before first status check then normal delay in loop
                    time.sleep(3)

                    poll_time = 0
                    while True:
                        self._tcex.log.debug(u'poll_time: {}'.format(poll_time))
                        if poll_time >= self._batch_poll_interval_max:
                            msg = u'Status check exceeded max poll time.'
                            self._tcex.log.error(msg)
                            self._tcex.message_tc(msg)
                            self._tcex.exit(1)

                        status = self.batch_status(batch_id)
                        if status.get('completed'):
                            if status.get('errors'):
                                if self._batch_halt_on_error:
                                    self._tcex.exit_code = 1
                                    halt = True
                                    # all indicator in chunk will be not_saved
                                    self._indicator_results['not_saved'].extend(
                                        [i.get('summary') for i in chunk])
                                    break
                                else:
                                    # all indicators were saved minus failed; not_save == failed
                                    self._indicator_results['not_saved'] = self._indicator_results.get(
                                        'failed', [])
                                    self._indicator_results['saved'].extend(
                                        [i.get('summary') for i in chunk
                                         if i.get('summary') not in self._indicator_results.get(
                                             'failed', [])])
                                    self._indicators_response.extend(
                                        [i for i in chunk
                                         if i.get('summary') not in self._indicator_results.get(
                                             'failed', [])])
                            else:
                                # all indicators were saved
                                self._indicator_results['saved'].extend(
                                    [i.get('summary') for i in chunk])
                                self._indicators_response.extend(chunk)
                            break  # no need to check status anymore

                        time.sleep(self._batch_poll_interval)
                        poll_time += self._batch_poll_interval
            else:
                self._tcex.log.warning('API request failed ({}).'.format(
                    results.get('response').text))
                # TODO: move this and above duplicate code to "if halt" below after validating logic
                self._tcex.exit_code = 1
                # all indicator in chunk will be not_saved
                self._indicator_results['not_saved'].extend([i.get('summary') for i in chunk])
                halt = True

            if halt:
                self._tcex.log.info(u'Halting on error.')
                break

    def _process_indicators_v2(self, owner):
        """Process batch indicators and write to ThreatConnect API using **/v2/indicators**.

        Args:
            owner (string):  The owner name for the indicator to be written.
        """
        self._indicator_results['not_saved'] = list(self._indicator_results.get('submitted', []))
        for i_data in self._indicators:
            resource = self._tcex.resource(i_data.get('type'))

            indicators = self._tcex.expand_indicators(i_data.get('summary'))
            try:
                i_value = [i for i in indicators if i is not None][0]
            except IndexError as e:
                err = u'Cannot proceed without an Indicator. ({})'.format(e)
                self._tcex.log.error(err)
                raise RuntimeError(e)

            body = resource.entity_body(indicators)
            if i_data.get('rating') is not None:
                body['rating'] = i_data.get('rating')
            if i_data.get('confidence') is not None:
                body['confidence'] = i_data.get('confidence')
            if i_data.get('size') is not None:
                body['size'] = i_data.get('size')
            if i_data.get('dns_active') is not None:
                body['dns_active'] = i_data.get('dns_active')
            if i_data.get('whois_active') is not None:
                body['whois_active'] = i_data.get('whois_active')
            if i_data.get('source') is not None:
                body['source'] = i_data.get('source')

            resource.body = json.dumps(body)
            resource.http_method = 'POST'
            resource.owner = owner
            resource.url = self._tcex.default_args.tc_api_path

            i_results = resource.request()

            # PUT file indicator since API does not work consistently for all indicator types
            if i_data.get('type') == 'File' and i_results.get('response').status_code == 400:
                if 'MD5' in i_results.get('response').text:
                    i_value = body.get('md5', i_value)
                elif 'SHA-1' in i_results.get('response').text:
                    i_value = body.get('sha1', i_value)
                elif 'SHA-256' in i_results.get('response').text:
                    i_value = body.get('sha256', i_value)
                resource.http_method = 'PUT'
                resource.indicator(i_value)
                i_results = resource.request()
                resource.http_method = 'POST'  # reset http method after PUT

            # Log error and continue
            if i_results.get('status') != 'Success':
                self._indicator_results['failed'].append(i_data.get('summary'))
                err = u'Failed adding indicator {} type {} ({}).'.format(
                    i_data.get('summary'), i_data.get('type'), i_results.get('response').text)
                self._tcex.log.error(err)

                # halt on error check
                if self._batch_halt_on_error:
                    self._tcex.log.info(u'Halt on error is enabled.')
                    self._tcex.exit_code = 1
                    break
                else:
                    self._tcex.exit_code = 3
                    continue

            # update indicator result list
            self._indicator_results['saved'].append(i_data.get('summary'))
            self._indicator_results['not_saved'].remove(i_data.get('summary'))

            # build indicator results output
            results_data = i_results.get('data')

            # Boolen for halt on error
            halt = False

            # Add attribute to Indicator
            for attribute in i_data.get('attribute', []):
                # process attributes
                if resource.custom:
                    resource.indicator(i_data.get('summary'))
                else:
                    resource.indicator(i_value)
                attribute_resource = resource.attributes()  # pivot to attribute
                attribute_resource.http_method = 'POST'
                attribute_resource.body = json.dumps(attribute)
                a_results = attribute_resource.request()

                if a_results.get('status') != 'Success':
                    err = u'Failed adding attribute type {} with value {} to indicator {}.'
                    err = err.format(
                        attribute.get('type'), attribute.get('value'), i_data.get('summary'))
                    self._tcex.log.error(err)

                    # halt on error check
                    if self._batch_halt_on_error:
                        self._tcex.log.info(u'Halt on error is enabled.')
                        self._tcex.exit_code = 1
                        halt = True
                        break
                    else:
                        self._tcex.exit_code = 3
                        continue
                results_data.setdefault('attribute', []).append(a_results.get('data'))
            if halt:
                break

            for tag in i_data.get('tag', []):
                # process attributes
                if resource.custom:
                    resource.indicator(i_data.get('summary'))
                else:
                    resource.indicator(i_value)
                tag_resource = resource.tags(tag.get('name'))
                tag_resource.http_method = 'POST'
                t_results = tag_resource.request()
                if t_results.get('status') != 'Success':
                    err = u'Failed adding tag {} to indicator {}.'.format(
                        tag.get('name'), i_data.get('summary'))
                    self._tcex.log.error(err)

                    # halt on error check
                    if self._batch_halt_on_error:
                        self._tcex.log.info(u'Halt on error is enabled.')
                        self._tcex.exit_code = 1
                        halt = True
                        break
                    else:
                        self._tcex.exit_code = 3
                        continue
                results_data.setdefault('tag', []).append(tag)
            if halt:
                break

            for group_id in i_data.get('associatedGroup', []):
                group_type = self.group_cache_type(group_id, owner)
                if group_type is None:
                    err = u'Could not get Group Type for Group ID {} in Owner "{}"'.format(
                        group_id, owner)
                    self._tcex.log.error(err)
                    self._tcex.exit_code = 3
                    continue

                if resource.custom:
                    resource.indicator(i_data.get('summary'))
                else:
                    resource.indicator(i_value)
                ar = self._tcex.resource(group_type)
                ar.resource_id(group_id)
                association_resource = resource.association_pivot(ar)
                association_resource.http_method = 'POST'
                a_results = association_resource.request()
                if a_results.get('status') != 'Success':
                    err = u'Failed association group id {} to indicator {}.'.format(
                        group_id, i_data.get('summary'))
                    self._tcex.log.error(err)

                    # halt on error check
                    if self._batch_halt_on_error:
                        self._tcex.log.info(u'Halt on error is enabled.')
                        self._tcex.exit_code = 1
                        halt = True
                        break
                    else:
                        self._tcex.exit_code = 3
                        continue
                results_data.setdefault('associatedGroup', []).append(group_id)
            if halt:
                break

            self._indicators_response.append(results_data)

    def association(self, associations):
        """Add association data to TcEx job.

        This method will add association data to the association list.

        .. Warning:: There is no validation of the data passed to this method.

        **Example Data** *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 2-5

            [{
              'association_value': '1.1.1.1',
              'association_type': 'Address',
              'resource_value': 'adversary-002',
              'resource_type': 'Adversary'
            },{
              'association_value': 'adversary-001',
              'association_type': 'Adversary',
              'resource_value': '1.1.1.1',
              'resource_type': 'Address'
            },{
              'association_value': 'adversary-001',
              'association_type': 'Adversary',
              'resource_value': 'threat-001',
              'resource_type': 'Threat',
            }]

        **Example Data for Creating Indicator-to-Indicator Associations**
        *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 2-6

            {
              "association_value": "ASN1234",
              "association_type": tcex.safe_rt("ASN"),
              "resource_value": "1.2.3.4",
              "resource_type": "Address",
              "custom_association_name": "ASN to Address"
            }

        Args:
            associations (dict | list): Dictionary or List containing
                association data
        """
        if isinstance(associations, list):
            self._associations.extend(associations)
        elif isinstance(associations, dict):
            self._associations.append(associations)

    def batch_action(self, action):
        """Set the default batch action for argument parser.

        Args:
            action (string): Set batch job action.
        """
        if action in ['Create', 'Delete']:
            self._batch_action = action

    def batch_chunk(self, chunk_size):
        """Set batch chunk_size for argument parser.

        Args:
            chunk_size (integer): Set batch job chunk size.
        """
        self._batch_chunk = chunk_size

    def batch_halt_on_error(self, halt_on_error):
        """Set batch halt on error boolean for argument parser.

        Args:
            halt_on_error (boolean): Boolean value for halt on error.
        """
        if isinstance(halt_on_error, bool):
            self._batch_halt_on_error = halt_on_error

    def batch_indicator_success(self):
        """Check completion for all batch jobs associated with this instance.

        Iterate over self._indicator_batch_ids set in
        :py:meth:`~tcex.tcex_job.TcExJob._process_indicators` and return the status of all job.

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 3

            {
              "status": "Success",
              "data": {
                "batchStatus": {
                  "id": 392,
                  "status": "Completed",
                  "errorCount": 1,
                  "successCount": 4,
                  "unprocessCount": 0
                }
              }
            }

        Returns:
            (dictionary): The status results from the Batch jobs.
        """

        status = {'success': 0, 'failure': 0, 'unprocessed': 0}

        resource = self._tcex.resource('Batch')
        resource.url = self._tcex.default_args.tc_api_path

        for batch_id in self._indicator_batch_ids:
            resource.batch_id(batch_id)
            results = resource.request()
            batch_status = results.get('data', {}).get('batchStatus', {})

            if results.get('status') == 'Success' and batch_status.get('status') == 'Completed':
                status['success'] += batch_status.get('successCount', 0)
                status['failure'] += batch_status.get('errorCount', 0)
                status['unprocessed'] += batch_status.get('unprocessCount', 0)

        return status

    def batch_poll_interval(self, interval):
        """Set batch polling interval for argument parser.

        Args:
            interval (integer): Seconds between polling.
        """
        if isinstance(interval, int):
            self._batch_poll_interval = interval

    def batch_poll_interval_max(self, interval_max):
        """Set batch polling interval max for argument parser.

        Args:
            interval_max (integer): Max seconds before timeout on batch.
        """
        if isinstance(interval_max, int):
            self._batch_poll_interval_max = interval_max

    def batch_status(self, batch_id):
        """Check the status of a batch job.

        This method will get the status of a batch job. Any errors returned from the batch status
        will be automatically logged.

        Critical errors are defined in the __init__ method in the self._batch_failures dictionary.
        If any of the critical errors are found the execution of the App will halt with a exit
        code of 1.  All other errors will set the status code to 3, but will not cause the execution
        to halt.

        Args:
            batch_id (int): Id of the batch job.

        Returns:
            (dict): A dictionary with status and error boolean.
        """
        status = {
            'completed': False,
            'errors': False
        }

        resource = self._tcex.resource('Batch')
        resource.url = self._tcex.default_args.tc_api_path
        resource.batch_id(batch_id)
        results = resource.request()

        self._tcex.log.info(u'batch id: {}'.format(batch_id))
        if results.get('status') == 'Success':
            if results['data']['status'] == 'Completed':
                status['completed'] = True

                status_msg = u'Batch Job completed, totals: '
                status_msg += u'succeeded: {}, '.format(results.get('data').get('successCount'))
                status_msg += u'failed: {}, '.format(results.get('data').get('errorCount'))
                status_msg += u'unprocessed: {}'.format(results.get('data').get('unprocessCount'))
                self._tcex.log.info(status_msg)

                if results['data']['errorCount'] > 0:
                    status['errors'] = True
                    self._tcex.exit_code = 3
                    time.sleep(2)

                    poll_time = 0
                    while True:
                        resource = self._tcex.resource('Batch')
                        resource.url = self._tcex.default_args.tc_api_path
                        resource.errors(batch_id)
                        error_results = resource.request()
                        if poll_time >= self._batch_poll_interval_max:
                            msg = u'Status check exceeded max poll time.'
                            self._tcex.log.error(msg)
                            self._tcex.message_tc(msg)
                            self._tcex.exit(1)
                        elif error_results.get('response').status_code == 200:
                            break
                        self._tcex.log.info(u'Error retrieve sleep ({} seconds)'.format(
                            self._batch_poll_interval))
                        time.sleep(self._batch_poll_interval)
                        poll_time += self._batch_poll_interval

                    try:
                        errors = json.loads(self._tcex.s(error_results.get('data')))
                    except TypeError as e:
                        err = u'Error loading Batch Error data ({})'.format(e)
                        self._tcex.log.error(err)
                        errors = []

                    for error in errors:
                        error_reason = error.get('errorReason')
                        error_source = error.get('errorSource')
                        # fix issue with API response
                        if error_source is not None and '  :  ' in error_source:
                            error_source = error.get('errorSource').replace('  ', ' ')
                        # errorSource won't always have an indicator per cblades
                        if error_source in self._indicator_results.get('submitted', []):
                            self._indicator_results['failed'].append(error_source)
                        for error_msg in self._batch_failures:
                            if re.findall(error_msg, error_reason):
                                # Critical Error
                                err = u'Batch Error: {}'.format(error)
                                self._tcex.message_tc(err)
                                self._tcex.log.error(err)
                                self._tcex.exit(1)
                        self._tcex.log.warn(u'Batch Error: {}'.format(error))
            else:
                self._tcex.log.debug(u'Batch Status: {}'.format(results.get('data').get('status')))

        return status

    def batch_write_type(self, write_type):
        """Set batch attributes write type for argument parser.

        Args:
            write_type (string): Type of Append or Replace.
        """
        if write_type in ['Append', 'Replace']:
            self._batch_write_type = write_type

    def file_occurrence(self, fo):
        """Add a file occurrence to job.

        Args:
            fo (dictionary): The file occurrence data.

        .. Warning:: There is no validation of the data passed to this method.

        **Example Data** *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 3

            {
                "date" : "2014-11-03T00:00:00-05:00",
                "fileName" : "win999301.dll",
                "hash": "BE7DE2F0CF48294400C714C9E28ECD01",
                "path" : "C:\\Windows\\System"
            }

        .. Note:: The hash in the example above is not posted in the body, but extracted to use
                  in the URI.
        """
        # POST /v2/indicators/files/BE7DE2F0CF48294400C714C9E28ECD01/fileOccurrences

        if isinstance(fo, list):
            self._file_occurrences.extend(fo)
        elif isinstance(fo, dict):
            self._file_occurrences.append(fo)

    def group(self, group):
        """Add group data to TcEx job.

        This method will accept a list or dictionary of formatted *Group* data and submit it to the
        API when the :py:meth:`~tcex.tcex_job.TcExJob.process` method is called.

        .. Warning:: There is no validation of the data passed to this method. Any duplicate group
                     name will be skipped.

        **Example Data** *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 9,14

            {
              'attribute': [
                {
                  'type': 'Description',
                  'value': 'Test Description'
                }
              ],
              'eventDate': '2015-03-7T00:00:00Z'
              'name': 'adversary-001',
              'tag': [{
                'name': 'Pop Star'
              }],
              'type': 'Adversary'
            }

        Args:
            group (dict | list): Dictionary or List containing group data.
        """
        if isinstance(group, dict):
            group = [group]

        for g in group:
            if g.get('name') not in [n.get('name') for n in self._groups]:
                self._group_results['submitted'].append(g.get('name'))
                self._groups.append(g)

    def group_association(self, associations):
        """Add group association data to TcEx job.

        This method will add group association data to the group association list.

        .. Warning:: There is no validation of the data passed to this method.

        **Example Data** *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 2-5

            {
              'group_name': 'adversary-001',
              'group_type': 'Adversary',
              'indicator': '1.1.1.1',
              'indicator_type': 'Address'
            }

        Args:
            associations (dict | list): Dictionary or List containing group
                association data.
        """
        if isinstance(associations, list):
            self._group_associations.extend(associations)
        elif isinstance(associations, dict):
            self._group_associations.append(associations)

    @property
    def group_association_len(self):
        """The current length of the group association list.

        Returns:
            (integer): The length of the group association list.
        """
        return len(self._group_associations)

    def group_cache(self, owner, resource_type):
        """Cache group data from the ThreatConnect Platform.

        The method will cache ThreatConnect group data by owner and type.

        **Cache Structure**
        ::

            Owner -> Group Type -> Group Name:Group Id

        Args:
            owner (string): The name of the ThreatConnect owner
            resource_type (string): The resource type name

        Returns:
            (dictionary): Dictionary of group resources
        """
        # already cached
        if owner in self._group_cache:
            if resource_type in self._group_cache.get(owner):
                return self._group_cache.get(owner).get(resource_type)

        self._tcex.log.info(u'Caching group type {} for owner {}'.format(resource_type, owner))
        self._group_cache.setdefault(owner, {})
        self._group_cache[owner].setdefault(resource_type, {})

        resource = self._tcex.resource(resource_type)
        resource.owner = owner
        resource.url = self._tcex.default_args.tc_api_path
        data = []
        for results in resource:
            if results.get('status') == 'Success':
                data += results.get('data')
            else:
                err = u'Failed retrieving result during pagination.'
                self._tcex.log.error(err)
                raise RuntimeError(err)

        for group in data:
            self._tcex.log.debug(u'cache - group name: ({})'.format(group.get('name')))
            dup = self._group_cache.get(owner, {}).get(resource_type, {}).get(group.get('name'))
            if dup is not None:
                warn = u'A duplicate Group name was found ({}). '
                warn += u'Duplicates are not supported in cache.'
                warn = warn.format(group.get('name'))
                self._tcex.log.warning(warn)
            self._group_cache[owner][resource_type][group['name']] = group.get('id')

        return self._group_cache.get(owner).get(resource_type)

    def group_id(self, name, owner, resource_type):
        """Get the group id from the group cache.

        Args:
            name(string): The name of the Group.
            owner (string): The TC Owner where the resource should be found.
            resource_type (string): The resource type name.

        Returns:
            (integer): The ID for the provided group name and owner.
        """
        self._tcex.log.debug(u'Retrieving {} ID for group "{}" in "{}".'.format(
            resource_type, name, owner
        ))
        if self._group_cache.get(owner, {}).get(resource_type) is None:
            self.group_cache(owner, resource_type)

        return self._group_cache.get(owner, {}).get(resource_type, {}).get(name)

    def group_cache_add(self, name, owner, resource_type, resource_id):
        """Add a group to the group cache.

        Args:
            name(string): The name of the Group.
            owner (string): The TC Owner where the resource should be found.
            resource_type (string): The resource type name.

        Returns:
            (integer): The ID for the provided group name and owner.
        """
        if self._group_cache.get(owner, {}).get(resource_type, {}) is None:
            self.group_cache(owner, resource_type)

        self._group_cache[owner][resource_type][name] = resource_id

    def group_cache_remove(self, name, owner, resource_type):
        """Remove a group from the group cache.

        Args:
            name(string): The name of the Group.
            owner (string): The TC Owner where the resource should be found.
            resource_type (string): The resource type name.

        Returns:
            (integer): The ID for the provided group name and owner.
        """
        self._tcex.log.debug(
            u'Removing {} ID for group "{}" in "{}".'.format(
                resource_type, name, owner))
        if self._group_cache.get(owner, {}).get(resource_type, {}).get(name) is not None:
            del self._group_cache[owner][resource_type][name]

    def group_cache_type(self, group_id, owner):
        """Get the group type for the provided group id.

        **Cache Structure**
        ::

            Owner -> Group Id:Group Type

        Args:
            group_id (string): The group id to lookup.
            owner (string): The TC Owner where the resource should be found.
            resource_type (string): The resource type name.

        Returns:
            (integer): The ID for the provided group name and owner.
        """
        group_id = int(group_id)
        if self._group_cache_id.get(owner) is None:

            self._tcex.log.info(u'Caching groups for owner {}'.format(owner))
            self._group_cache_id.setdefault(owner, {})

            resource = self._tcex.resource('Group')
            resource.owner = owner
            resource.url = self._tcex.default_args.tc_api_path
            for results in resource:
                if results['status'] == 'Success':
                    for group in results.get('data'):
                        self._group_cache_id[owner][int(group.get('id'))] = group.get('type')
                else:
                    err = u'Failed retrieving result during pagination.'
                    self._tcex.log.error(err)
                    raise RuntimeError(err)

        return self._group_cache_id.get(owner, {}).get(group_id)

    @property
    def group_len(self):
        """The current length of the group list.

        Returns:
            (integer): The length of the group list.
        """
        return len(self._groups)

    @property
    def group_results(self):
        """Result dictionary of failed, saved, not_saved, and submitted groups.

        Returns:
            (dictionary): Dictionary of group names for each status.
        """
        return self._group_results

    def indicator(self, indicator):
        """Add indicator data to TcEx job.

        This method will add indicator data to this TcEx job to be submitted via batch import.

        .. Warning:: There is no validation of the data passed to this method. Any duplicate
                     indicator will be skipped.

        **Example Data** *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 14,19

            {
              'associatedGroup': [
                '1',
                '8'
              ],
              'attribute': [
                {
                  'type': 'Attribute Name',
                  'value': 'Description'
                }
              ],
              'confidence': 5,
              'rating': 3.2,
              'summary': '1.1.1.1',
              'tag': [
                {
                  'name': 'APT',
                },
                {
                  'name': 'CrimeWare',
                }
              ],
              'type': 'Address'
            }

        Args:
            indicator (dict | list): Dictionary or List containing indicator
                data.
        """
        if isinstance(indicator, dict):
            indicator = [indicator]

        for i in indicator:  # Not extending now since we need to build submitted list
            # verify indicator is not a duplicate before adding
            if i.get('summary') not in self._indicators:
                self._indicator_results['submitted'].append(i.get('summary'))
                self._indicators.append(i)

    @property
    def unprocessed_indicators(self):
        """Return indicators (unprocessed).

        Returns:
            (list): The list of unprocessed indicator.
        """
        return self._indicators

    @unprocessed_indicators.setter
    def unprocessed_indicators(self, value):
        """Set indicators list.

        Args:
            value (list): List of indicators

        """
        self._indicators = value

    @property
    def indicator_data(self):
        """Return the current indicator list.

        Returns:
            (list): The indicator list.
        """
        return self._indicators_response

    @property
    def indicator_len(self):
        """The current length of the indicator list.

        Returns:
            (integer): The length of the indicator list
        """
        return len(self._indicators)

    @property
    def indicator_results(self):
        """Result dictionary of failed, saved, not_saved, and submitted indicators.

        Returns:
            (dictionary): Dictionary of indicator values for each status.
        """
        return self._indicator_results

    def process(self, owner, indicator_batch=True, group_action='skip'):
        """Process all groups, indicator data, and associations.

        Process each of the supported data types for this job, in the following
        order (left to right):

            groups > indicators > file occurrences > group associations > associations

        Args:
            owner (string): The owner name for the data to be written.
            indicator_batch (bool): If true use the Batch Api otherwise use `/v2` REST API.
            group_action (string): The action to use on group create (duplicate, replace, skip).
        """
        if self._groups:
            self._tcex.log.info(u'Processing Groups')
            self._process_groups(owner, group_action.lower())

        if self._indicators:
            # update indicator data replacing group stub with group ids
            self._tcex.log.info(u'Processing Indicators')
            self._indicators_replace_stub(owner)
            self._process_indicators(owner, indicator_batch)

        if self._file_occurrences:
            self._tcex.log.info(u'Processing File Occurrences')
            self._process_file_occurrences(owner)

        if self._group_associations:
            self._tcex.log.info(u'Processing Group Associations')
            self._process_group_association(owner)

        if self._associations:
            self._tcex.log.info(u'Processing Associations')
            self._process_association(owner)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import cleaner


def adversary_create(batch_job):
    batch_job.add_group({
        'name': 'adversary-003',
        'type': 'Adversary',
        'xid': cleaner._create_xid('Adversary', 'adversary-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })


def campaign_create(batch_job):
    batch_job.add_group({
        'name': 'campaign-003',
        'type': 'Campaign',
        'xid': cleaner._create_xid('Campaign', 'campaign-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })


def document_create(batch_job):
    batch_job.add_group({
        'name': 'document-003',
        'fileName': 'test.txt',
        'fileContent': 'Document content',
        'type': 'Document',
        'xid': cleaner._create_xid('Document', 'document-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })


def document_malware_create(batch_job):
    batch_job.add_group({
        'name': 'document-malware-003',
        'fileName': 'test.txt',
        'fileContent': 'Document content',
        'type': 'Document',
        'malware': True,
        'xid': cleaner._create_xid('Document', 'document-malware-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })


def email_create(batch_job):
    batch_job.add_group({
        'name': 'email-003',
        'type': 'Email',
        'xid': cleaner._create_xid('Email', 'email-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }],
        'subject': 'Greetings!',
        'header': 'This is just a test',
        'body': 'This is just a test'
    })


def event_create(batch_job):
    batch_job.add_group({
        'name': 'event-003',
        'type': 'Event',
        'xid': cleaner._create_xid('Event', 'event-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })


def incident_create(batch_job):
    batch_job.add_group({
        'name': 'incident-003',
        'type': 'Incident',
        'xid': cleaner._create_xid('Incident', 'incident-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })


def intrusion_set_create(batch_job):
    batch_job.add_group({
        'name': 'intrusion_set-003',
        'type': 'Intrusion Set',
        'xid': cleaner._create_xid('Intrusion Set', 'intrusion_set-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })


def report_create(batch_job):
    batch_job.add_group({
        'name': 'report-003',
        'fileName': 'test.txt',
        'fileContent': 'Report content',
        'type': 'Report',
        'xid': cleaner._create_xid('Report', 'report-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })


def signature_create(batch_job):
    batch_job.add_group({
        'name': 'signature-003',
        'fileName': 'test.snort',
        'fileType': 'Snort',
        'fileText': 'Signature content',
        'type': 'Signature',
        'xid': cleaner._create_xid('Signature', 'signature-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })


def threat_create(batch_job):
    batch_job.add_group({
        'name': 'threat-003',
        'type': 'Threat',
        'xid': cleaner._create_xid('Threat', 'threat-003'),
        'attribute': [{
            "displayed": True,
            "type": "Description",
            "value": "Example Description"
        }],
        'tag': [{
            'name': 'Example Tag'
        }]
    })



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


def address_create(batch_job):
    address = batch_job.address('123.124.125.126', '5.0', '100')
    address.attribute('Description', 'Example Description', True)
    address.tag('Example Tag')
    address.security_label('TLP Green')


def email_address_create(batch_job):
    email_address = batch_job.email_address('bad@dfadsfaddsfa.com', '5.0', '100')
    email_address.attribute('Description', 'Example Description', True)
    email_address.tag('Example Tag')
    email_address.security_label('TLP Green')


def file_create(batch_job):
    file = batch_job.file(md5='a'*32, sha1='b'*40, sha256='c'*64, rating='5.0', confidence='100')
    file.attribute('Description', 'Example Description', True)
    file.tag('Example Tag')
    file.security_label('TLP Green')


def host_create(batch_job):
    host = batch_job.host('dfadsfaddsfa.com', rating='5.0', confidence='100')
    host.attribute('Description', 'Example Description', True)
    host.tag('Example Tag')
    host.security_label('TLP Green')


def url_create(batch_job):
    url = batch_job.url('https://dfadsfaddsfa.com/index.html', '5.0', '100')
    url.attribute('Description', 'Example Description', True)
    url.tag('Example Tag')
    url.security_label('TLP Green')



# -*- coding: utf-8 -*-
"""ThreatConnect REST API Authentication"""
import base64
import hashlib
import hmac
import logging
import sys
import time

from requests import (auth, exceptions, get)


class TcExAuth(auth.AuthBase):
    """ThreatConnect Authorization Class"""
    def __init__(self, logger=None):
        """Initialize Class Properties"""
        self.log = self._logger()
        if logger is not None:
            self.log = logger
        # renewing state for token auth
        self.renewing = False

    @staticmethod
    def _logger():
        """Initialize basic stream logger."""
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        logger.addHandler(ch)
        return logger

    def __call__(self, r):
        """Override of parent __call__ method."""
        pass


class TcExHmacAuth(TcExAuth):
    """ThreatConnect HMAC Authorization"""
    def __init__(self, access_id, secret_key, logger=None):
        """Initialize the Class properties."""
        super(TcExHmacAuth, self).__init__(logger)
        self._access_id = access_id
        self._secret_key = secret_key

    def __call__(self, r):
        """Override of parent __call__ method."""
        timestamp = int(time.time())
        signature = '{}:{}:{}'.format(r.path_url, r.method, timestamp)
        hmac_signature = hmac.new(
            self._secret_key.encode(), signature.encode(), digestmod=hashlib.sha256).digest()
        authorization = 'TC {}:{}'.format(
            self._access_id, base64.b64encode(hmac_signature).decode())
        r.headers['Authorization'] = authorization
        r.headers['Timestamp'] = timestamp
        return r


class TcExTokenAuth(TcExAuth):
    """ThreatConnect Token Authorization"""
    def __init__(self, session, token, token_expiration, token_url, logger=None):
        """Initialize Class Properties."""
        super(TcExTokenAuth, self).__init__(logger)
        self._token = token
        self._token_expiration = int(token_expiration)
        self._token_url = token_url
        self._session = session  # use same value as session for verify

    def _renew_token(self, retry=True):
        """Renew expired ThreatConnect Token."""
        self.renewing = True
        self.log.info('Renewing ThreatConnect Token')
        self.log.info('Current Token Expiration: {}'.format(self._token_expiration))
        try:
            params = {'expiredToken': self._token}
            url = '{}/appAuth'.format(self._token_url)
            r = get(url, params=params, verify=self._session.verify)
            if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                if (r.status_code == 401 and
                        'application/json' in r.headers.get('content-type', '') and
                        'Retry token is invalid' in r.json().get('message')):
                    # TODO: remove this once token renewal issue is fixed
                    self.log.error('params: {}'.format(params))
                    self.log.error('url: {}'.format(r.url))

                    # log failure
                    err_reason = r.text or r.reason
                    err_msg = 'Token Retry Error. API status code: {}, API message: {}.'
                    raise RuntimeError(1042, err_msg.format(r.status_code, err_reason))
                elif retry:
                    warn_msg = 'Token Retry Error. API status code: {}, API message: {}.'
                    self.log.warning(warn_msg.format(r.status_code, r.text))
                    # delay and retry token renewal
                    time.sleep(15)
                    self._renew_token(False)
                else:
                    err_reason = r.text or r.reason
                    err_msg = 'Token Retry Error. API status code: {}, API message: {}.'
                    raise RuntimeError(1042, err_msg.format(r.status_code, err_reason))
            data = r.json()
            if retry and (data.get('apiToken') is None or data.get('apiTokenExpires') is None):
                # add retry logic to handle case if the token renewal doesn't return valid data
                warn_msg = 'Token Retry Error: no values for apiToken or apiTokenExpires ({}).'
                self.log.warning(warn_msg.format(r.text))
                self._renew_token(False)
            else:
                self._token = data.get('apiToken')
                self._token_expiration = int(data.get('apiTokenExpires'))
                self.log.info('New Token Expiration: {}'.format(self._token_expiration))
            self.renewing = False
        except exceptions.SSLError:
            self.log.error(u'SSL Error during token renewal.')
            self.renewing = False

    def __call__(self, r):
        """Override of parent __call__ method."""
        window_padding = 60  # pad renewal by 60 seconds
        current_time = int(time.time()) + window_padding
        if self._token_expiration < current_time:
            self._renew_token()
        r.headers['Authorization'] = 'TC-Token {}'.format(self._token)
        return r



# -*- coding: utf-8 -*-
""" TcEx Error Codes """

class TcExErrorCodes(object):
    """TcEx Framework Error Codes."""

    @property
    def errors(self):
        """TcEx defined error codes and messages.

        .. note:: RuntimeErrors with a code of >= 1000 are considered critical.  Those < 1000
            are considered warning or errors and are up to the developer to determine the appropriate
            behavior.
        """
        return {
            # tcex general errors
            100: 'Generic error. See log for more details ({}).',
            105: 'Required Module is not installed ({}).',
            200: 'Failed retrieving Custom Indicator Associations types from API ({}).',
            210: 'Failure during token renewal ({}).',
            215: 'HMAC authorization requires a PreparedRequest Object.',
            220: 'Failed retrieving indicator types from API ({}).',
            # tcex resource
            300: 'Failed retrieving Bulk JSON ({}).',
            305: 'An invalid action/association name ({}) was provided.',
            350: 'Data Store request failed. API status code: {}, API message: {}.',
            # batch v2: 500-600
            520: 'File Occurrences can only be added to a File. Current type: {}.',
            540: 'Failed polling batch status ({}).',
            545: 'Failed polling batch status. API status code: {}, API message: {}.',
            550: 'Batch status check reached timeout ({} seconds).',
            560: 'Failed retrieving batch errors ({}).',
            580: 'Failed posting file data ({}).',
            585: 'Failed posting file data. API status code: {}, API message: {}.',
            590: 'No hash values provided.',
            # metrics
            700: 'Failed to create metric. API status code: {}, API message: {}.',
            705: 'Error while finding metric by name. API status code: {}, API message: {}.',
            710: 'Failed to add metric data. API status code: {}, API message: {}.',
            715: 'No metric ID found for "{}".',
            # batch v2 critical: 1500-1600
            1500: 'Critical batch error ({}).',
            1505: 'Failed submitting batch job requests ({}).',
            1510: 'Failed submitting batch job requests. API status code: {}, API message: {}.',
            1520: 'Failed submitting batch data ({}).',
            1525: 'Failed submitting batch data. API status code: {}, API message: {}.',
        }

    def message(self, code):
        """Return the error message.

        Args:
            code (integer): The error code integer.

        Returns:
            (string): The error message.
        """
        return self.errors.get(code)



# -*- coding: utf-8 -*-
"""TcEx Framework KeyValue Module"""
from builtins import str
try:
    from urllib import quote  # Python 2
except ImportError:
    from urllib.parse import quote  # Python 3


class TcExKeyValue(object):
    """Update Redis via ThreatConnect API Wrapper"""

    def __init__(self, tcex):
        """Initialize the Class properties.

        Args:
            tcex (object): Instance of TcEx.
            rhash (string): The REDIS hash.
        """
        self.tcex = tcex

    def create(self, key, value):
        """Create key/value pair in remote KV store.

        Args:
            key (string): The key to create in remote KV store.
            value (any): The value to store in remote KV store.

        Returns:
            (string): The response from the API call.
        """
        key = quote(key, safe='~')
        headers = {'content-type': 'application/octet-stream'}
        url = '/internal/playbooks/keyValue/{}'.format(key)
        r = self.tcex.session.put(url, data=value, headers=headers)
        return r.content

    # def delete(self, key):
    #     """Delete is not supported in API Wrapper"""
    #     return None

    def read(self, key):
        """Read data from remote KV store for the provided key.

        Args:
            key (string): The key to read in remote KV store.

        Returns:
            (any): The response data from the remote KV store.
        """
        key = quote(key, safe='~')
        url = '/internal/playbooks/keyValue/{}'.format(key)
        r = self.tcex.session.get(url)
        data = r.content
        if data is not None and not isinstance(data, str):
            data = str(r.content, 'utf-8')
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Delete indicators and groups which have been created for the tests."""

import hashlib
import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))
import validator


def _create_xid(type_, name):
    # if given a file indicator, make sure the name is based on the first hash
    if type_ == 'file':
        name = name.split(' : ')[0]
    xid_string = '{}-{}'.format(type_, name)
    hash_object = hashlib.sha256(xid_string.encode('utf-8'))
    return hash_object.hexdigest()


def _delete_groups(tcex):
    batch = tcex.batch(tcex.args.api_default_org, action='Delete')
    groups = validator.get_groups(tcex)
    for group in groups:
        group['xid'] = _create_xid(group['type'], group['name'])
        batch.add_group(group)
    batch.submit_all()


def _delete_indicators(tcex):
    # TODO: implement
    batch = tcex.batch(tcex.args.api_default_org, action='Delete')
    indicators = validator.get_indicators(tcex)
    for indicator in indicators:
        indicator['xid'] = _create_xid(indicator['type'], indicator['summary'])
        batch.add_indicator(indicator)
    batch.submit_all()


def clean(tcex):
    """Delete all indicators and groups in the source."""
    _delete_groups(tcex)
    _delete_indicators(tcex)



# -*- coding: utf-8 -*-
""" TcEx Framework """
from builtins import str
import inspect
import json
import logging
import platform
import os
import re
import signal
import sys
import time
try:
    from urllib import quote  # Python 2
except ImportError:
    from urllib.parse import quote  # Python 3

from .tcex_argparser import TcExArgParser


class TcEx(object):
    """Provides basic functionality for all types of TxEx Apps."""

    def __init__(self):
        """Initialize Class Properties."""
        # catch interupt signals
        signal.signal(signal.SIGINT, self._signal_handler)
        if platform.system() != 'Windows':
            signal.signal(signal.SIGHUP, self._signal_handler)
        signal.signal(signal.SIGTERM, self._signal_handler)
        # init logger
        self.log = logging.getLogger('tcex')
        self._logger_stream()
        # Property defaults
        self._error_codes = None
        self._exit_code = 0
        self._default_args = None
        self._install_json = None
        self._install_json_params = {}
        self._indicator_associations_types_data = {}
        self._indicator_types = []
        self._indicator_types_data = {}
        self._jobs = None
        self._playbook = None
        self._session = None
        self._utils = None

        # NOTE: odd issue where args is not updating properly
        self._tc_token = None
        self._tc_token_expires = None

        # Parser
        self._parsed = False
        self.parser = TcExArgParser()

        # Ensure App is not already running with same session id.
        # self._singular()

        # NOTE: odd issue where args is not updating properly
        if self.default_args.tc_token is not None:
            self._tc_token = self.default_args.tc_token
        if self.default_args.tc_token_expires is not None:
            self._tc_token_expires = self.default_args.tc_token_expires

        # Log system and App data
        self._log()

        # include resources module
        self._resources()

    def _association_types(self):
        """Retrieve Custom Indicator Associations types from the ThreatConnect API."""
        # Dynamically create custom indicator class
        r = self.session.get('/v2/types/associationTypes')

        # check for bad status code and response that is not JSON
        if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
            warn = u'Custom Indicators Associations are not supported.'
            self.log.warning(warn)
            return

        # validate successful API results
        data = r.json()
        if data.get('status') != 'Success':
            warn = u'Bad Status: Custom Indicators Associations are not supported.'
            self.log.warning(warn)
            return

        try:
            # Association Type Name is not a unique value at this time, but should be.
            for association in data.get('data', {}).get('associationType', []):
                self._indicator_associations_types_data[association.get('name')] = association
        except Exception as e:
            self.handle_error(200, [e])

    def _authorization_token_renew(self):
        """Method for handling token authorization to ThreatConnect API.

        This method will automatically renew the ThreatConnect token if it has expired.

        Returns:
            (dictionary): An dictionary containing the header values for authorization to
                          ThreatConnect.
        """
        authorization = 'TC-Token {}'.format(self._tc_token)

        window_padding = 15  # bcs - possible configuration option
        current_time = int(time.time()) + window_padding
        if self._tc_token_expires < current_time:
            # Renew Token
            request = self.request()
            request.add_payload('expiredToken', self._tc_token)
            request.url = '{}/appAuth'.format(self.default_args.tc_api_path)
            r = request.send()
            if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                self.handle_error(210, [r.text])
            data = r.json()
            if data['success']:
                self.log.info(u'Expired API token has been renewed.')
                self._tc_token = data['apiToken']  # remove str() due to newstr issue
                self._tc_token_expires = int(data['apiTokenExpires'])
                authorization = 'TC-Token {}'.format(data['apiToken'])
            else:
                self.handle_error(210, [r.text])
        return {'Authorization': authorization}

    def _load_secure_params(self):
        """Load secure params from the API.

        # API Response:

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1

            {
                "inputs":
                    {
                        "tc_playbook_db_type": "Redis",
                        "fail_on_error": true,
                        "api_default_org": "TCI"
                    }
            }

        Returns:
            dict: Parameters ("inputs") from the TC API.
        """
        self.log.info('Loading secure params.')
        # Retrieve secure params and inject them into sys.argv
        r = self.session.get('/internal/job/execution/parameters')

        # check for bad status code and response that is not JSON
        if not r.ok or r.headers.get('content-type') != 'application/json':
            err = r.text or r.reason
            self.exit(1, 'Error retrieving secure params from API ({}).'.format(err))

        # return secure params
        return r.json().get('inputs', {})

    def _log(self):
        """Send System and App data to logs."""
        self._log_platform()
        self._log_app_data()
        self._log_python_version()
        self._log_tcex_version()
        self._log_tc_proxy()

    def _log_app_data(self):
        """Log the App data information."""
        # Best Effort
        if self.install_json:
            app_commit_hash = self.install_json.get('commitHash')
            app_features = ','.join(self.install_json.get('features', []))
            app_min_ver = self.install_json.get('minServerVersion', 'N/A')
            app_name = self.install_json.get('displayName')
            app_runtime_level = self.install_json.get('runtimeLevel')
            app_version = self.install_json.get('programVersion')

            self.log.info(u'App Name: {}'.format(app_name))
            if app_features:
                self.log.info(u'App Features: {}'.format(app_features))
            self.log.info(u'App Minimum ThreatConnect Version: {}'.format(app_min_ver))
            self.log.info(u'App Runtime Level: {}'.format(app_runtime_level))
            self.log.info(u'App Version: {}'.format(app_version))
            if app_commit_hash is not None:
                self.log.info(u'App Commit Hash: {}'.format(app_commit_hash))

    def _log_platform(self):
        """Log the current Platform."""
        from platform import platform
        self.log.info(u'Platform: {}'.format(platform()))

    def _log_python_version(self):
        """Log the current Python version."""
        self.log.info(u'Python Version: {}.{}.{}'.format(
            sys.version_info.major, sys.version_info.minor, sys.version_info.micro))

    def _log_tc_proxy(self):
        """Log the proxy settings."""
        if self.default_args.tc_proxy_tc:
            self.log.info(u'Proxy Server (TC): {}:{}.'.format(
                self.default_args.tc_proxy_host, self.default_args.tc_proxy_port))

    def _log_tcex_version(self):
        """Log the current TcEx version number."""
        self.log.info(u'TcEx Version: {}'.format(__import__(__name__).__version__))

    def _logger(self):
        """Create TcEx app logger instance.

        The logger is accessible via the ``tc.log.<level>`` call.

        **Logging examples**

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            tcex.log.debug('logging debug')
            tcex.log.info('logging info')
            tcex.log.warning('logging warning')
            tcex.log.error('logging error')

        Args:
            stream_only (bool, default:False): If True only the Stream handler will be enabled.

        Returns:
            logger: An instance of logging
        """
        level = logging.INFO
        self.log.setLevel(level)

        # clear all handlers
        self.log.handlers = []

        # update logging level
        if self.default_args.logging is not None:
            level = self._logger_levels[self.default_args.logging]
        elif self.default_args.tc_log_level is not None:
            level = self._logger_levels[self.default_args.tc_log_level]
        self.log.setLevel(level)

        # add file handler if not already added
        if self.default_args.tc_log_path:
            self._logger_fh()

        # add api handler if not already added
        if self.default_args.tc_token is not None and self.default_args.tc_log_to_api:
            self._logger_api()

        self.log.info('Logging Level: {}'.format(logging.getLevelName(level)))

    def _logger_api(self):
        """Add API logging handler."""
        from .tcex_logger import TcExLogHandler, TcExLogFormatter
        api = TcExLogHandler(self.session)
        api.set_name('api')
        api.setLevel(logging.DEBUG)
        api.setFormatter(TcExLogFormatter())
        self.log.addHandler(api)

    def _logger_fh(self):
        """Add File logging handler."""
        logfile = os.path.join(self.default_args.tc_log_path, self.default_args.tc_log_file)
        fh = logging.FileHandler(logfile)
        fh.set_name('fh')
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(self._logger_formatter)
        self.log.addHandler(fh)

    @property
    def _logger_formatter(self):
        """Return log formatter."""
        tx_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s '
        tx_format += '(%(funcName)s:%(lineno)d)'
        return logging.Formatter(tx_format)

    @property
    def _logger_levels(self):
        """Return log levels."""
        return {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL
        }

    def _logger_stream(self):
        """Add stream logging handler."""
        sh = logging.StreamHandler()
        sh.set_name('sh')
        sh.setLevel(logging.INFO)
        sh.setFormatter(self._logger_formatter)
        self.log.addHandler(sh)

    def _resources(self, custom_indicators=False):
        """Initialize the resource module.

        This method will make a request to the ThreatConnect API to dynamically
        build classes to support custom Indicators.  All other resources are available
        via this class.

        .. Note:: Resource Classes can be accessed using ``tcex.resources.<Class>`` or using
                  tcex.resource('<resource name>').
        """
        from importlib import import_module
        # create resource object
        self.resources = import_module('tcex.tcex_resources')

        if custom_indicators:
            self.log.info('Loading custom indicator types.')
            # Retrieve all indicator types from the API
            r = self.session.get('/v2/types/indicatorTypes')

            # check for bad status code and response that is not JSON
            if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                warn = u'Custom Indicators are not supported ({}).'.format(r.text)
                self.log.warning(warn)
                return
            response = r.json()
            if response.get('status') != 'Success':
                warn = u'Bad Status: Custom Indicators are not supported ({}).'.format(r.text)
                self.log.warning(warn)
                return

            try:
                # Dynamically create custom indicator class
                data = response.get('data', {}).get('indicatorType', [])
                for entry in data:
                    name = self.safe_rt(entry.get('name'))
                    # temp fix for API issue where boolean are returned as strings
                    entry['custom'] = self.utils.to_bool(entry.get('custom'))
                    entry['parsable'] = self.utils.to_bool(entry.get('parsable'))
                    self._indicator_types.append(u'{}'.format(entry.get('name')))
                    self._indicator_types_data[entry.get('name')] = entry
                    if not entry['custom']:
                        continue

                    # Custom Indicator have 3 values. Only add the value if it is set.
                    value_fields = []
                    if entry.get('value1Label'):
                        value_fields.append(entry.get('value1Label'))
                    if entry.get('value2Label'):
                        value_fields.append(entry.get('value2Label'))
                    if entry.get('value3Label'):
                        value_fields.append(entry.get('value3Label'))

                    # get instance of Indicator Class
                    i = self.resources.Indicator(self)
                    custom = {
                        '_api_branch': entry['apiBranch'],
                        '_api_entity': entry['apiEntity'],
                        '_api_uri': '{}/{}'.format(i.api_branch, entry['apiBranch']),
                        '_case_preference': entry['casePreference'],
                        '_custom': entry['custom'],
                        '_name': name,
                        '_parsable': entry['parsable'],
                        '_request_entity': entry['apiEntity'],
                        '_request_uri': '{}/{}'.format(i.api_branch, entry['apiBranch']),
                        '_status_codes': {
                            'DELETE': [200],
                            'GET': [200],
                            'POST': [200, 201],
                            'PUT': [200]
                        },
                        '_value_fields': value_fields
                    }
                    # Call custom indicator class factory
                    setattr(self.resources, name, self.resources.class_factory(
                        name, self.resources.Indicator, custom))
            except Exception as e:
                self.handle_error(220, [e])

    # def _singular(self):
    #     """There can be only one. Validate the App doesn't already have a lock file."""
    #     lock_file = os.path.join(self.default_args.tc_temp_path, 'app.lock')
    #     if os.path.isfile(lock_file):
    #         # append pid to lock file and exit
    #         self._singular_lock(lock_file)
    #         self.exit(0)
    #     # create new lockfile
    #     self._singular_lock(lock_file)

    # @staticmethod
    # def _singular_lock(lock_file):
    #     """Create or update the lock file."""
    #     with open(lock_file, 'a') as fh:
    #         fh.write('timestamp: {}\n'.format(time.time()))
    #         fh.write('pid: {}\n'.format(os.getpid()))

    def _signal_handler(self, signal_interupt, frame):
        """Handle singal interrupt."""
        call_file = os.path.basename(inspect.stack()[1][0].f_code.co_filename)
        call_module = inspect.stack()[1][0].f_globals['__name__'].lstrip('Functions.')
        call_line = inspect.stack()[1][0].f_lineno
        self.log.error(
            'App interrupted - file: {}, method: {}, line: {}.'.format(
                call_file, call_module, call_line))
        if signal_interupt in (2, 15):
            self.exit(1, 'The App received an interrupt signal and will now exit.')

    def _unknown_args(self, args):
        """Log argparser unknown arguments.

        Args:
            args (list): List of unknown arguments
        """
        for u in args:
            self.log.warning(u'Unsupported arg found ({}).'.format(u))

    @property
    def utils(self):
        """Include the Utils module.

        .. Note:: Utils methods can be accessed using ``tcex.utils.<method>``.
        """
        if self._utils is None:
            from .tcex_utils import TcExUtils
            self._utils = TcExUtils(self)
        return self._utils

    @property
    def args(self):
        """The parsed args from argparser

        .. Note:: Accessing args should only be done directly in the App.

        Returns:
            (namespace): ArgParser parsed arguments
        """

        if not self._parsed:
            self._default_args, unknown = self.parser.parse_known_args()
            self.results_tc_args()  # for local testing only
            self._parsed = True

            # log unknown arguments only once
            self._unknown_args(unknown)

        return self._default_args

    def authorization(self, request_prepped):
        """A method to handle the different methods of authenticating to the ThreatConnect API.

        **Token Based Authorization**::

            {'Authorization': authorization}

        **HMAC Based Authorization**::

            {
                'Authorization': authorization,
                'Timestamp': <unix timestamp>
            }

           http://docs.python-requests.org/en/master/api/#requests.Session.prepare_request.

        Args:
            request_prepped (object): A instance of Python Request module requests.
                                        PreparedRequest.
        Returns:
            (dictionary): An dictionary containing the header values for authorization to
                          ThreatConnect.
        """
        authorization = None

        if self._tc_token is not None:
            authorization = {'Authorization': 'TC-Token {}'.format(self._tc_token)}
            if self._tc_token_expires is not None:
                authorization = self._authorization_token_renew()
        elif (self.default_args.api_access_id is not None and
              self.default_args.api_secret_key is not None):
            authorization = self.authorization_hmac(request_prepped)

        return authorization

    def authorization_hmac(self, request_prepped):
        """Method for handling HMAC authorization to ThreatConnect API.

           http://docs.python-requests.org/en/master/api/#requests.Session.prepare_request.

        Args:
            request_prepped (object): A instance of Python Request prepped requests.
                                        PreparedRequest.
        Returns:
            (dictionary): An dictionary containing the header values for authorization to
                          ThreatConnect.
        """
        import base64
        import hashlib
        import hmac
        if request_prepped is None:
            self.handle_error(215, [])

        timestamp = int(time.time())
        signature = '{}:{}:{}'.format(
            request_prepped.path_url, request_prepped.method, timestamp)
        hmac_signature = hmac.new(
            self.default_args.api_secret_key.strip('\'').encode(), signature.encode(),
            digestmod=hashlib.sha256).digest()
        authorization = 'TC {}:{}'.format(
            self.default_args.api_access_id, base64.b64encode(hmac_signature).decode())
        return {
            'Authorization': authorization,
            'Timestamp': str(timestamp)
        }

    def batch(self, owner, action=None, attribute_write_type=None, halt_on_error=False,
              playbook_triggers_enabled=None):
        """Return instance of Batch"""
        from .tcex_batch_v2 import TcExBatch
        return TcExBatch(self, owner, action, attribute_write_type, halt_on_error, playbook_triggers_enabled)

    def bulk_enabled(self, owner=None, api_path=None, authorization=None):
        """[Deprecated] Check if bulk indicators is enabled for owner.

        .. warning:: This method is deprecated and will be removed in TcEx version 0.9.0.

        Using the TC API validate that bulk indicator download is enabled and
        has successfully run for the provided owner.

        Args:
            owner (Optional [string]): Owner name to check.
            api_path (Optional [string]): The url to the ThreatConnect API.
            authorization (Optional [string]): The authorization header value.

        Returns:
            (boolean): True if bulk indicator download is enabled and has run
        """
        self.log.warning('This App is using a deprecated method (bulk_enabled).')
        if api_path is None:
            api_path = self.default_args.tc_api_path

        # Dynamically create custom indicator class
        request = self.request()
        if authorization is not None:
            request.authorization = authorization
        if owner is not None:
            request.owner = owner
        request.url = '{}/v2/indicators/bulk'.format(api_path)

        r = request.send()
        if r.ok and 'application/json' in r.headers.get('content-type', ''):
            data = r.json()
            if data.get('status') == 'Success':
                if (data.get('data', {}).get('bulkStatus', {}).get('jsonEnabled') and
                        data.get('data').get('bulkStatus', {}).get('lastRun') is not None):
                    return True
        return False

    def data_filter(self, data):
        """Return an instance of the Data Filter Class.

        A simple helper module to filter results from ThreatConnect API or other data
        source.  For example if results need to be filtered by an unsupported field the module
        allows you to pass the data array/list in and specify one or more filters to get just the
        results required.

        Args:
            data (list): The list of dictionary structure to filter.

        Returns:
            (object): An instance of DataFilter Class
        """
        try:
            from .tcex_data_filter import DataFilter
            return DataFilter(self, data)
        except ImportError as e:
            warn = u'Required Module is not installed ({}).'.format(e)
            self.log.warning(warn)

    @property
    def default_args(self):
        """Parse args and return default args."""
        if self._default_args is None:
            self._default_args, unknown = self.parser.parse_known_args()
            # reinitialize logger with new log level and api settings
            self._logger()
            if self._default_args.tc_aot_enabled:
                # block for AOT message and get params
                params = self.playbook.aot_blpop()
                self.inject_params(params)
            elif self.default_args.tc_secure_params:
                # inject secure params from API
                params = self._load_secure_params()
                self.inject_params(params)
        return self._default_args

    @property
    def error_codes(self):
        """ThreatConnect error codes."""
        if self._error_codes is None:
            from .tcex_error_codes import TcExErrorCodes
            self._error_codes = TcExErrorCodes()
        return self._error_codes

    def exit(self, code=None, msg=None):
        """Application exit method with proper exit code

        The method will run the Python standard sys.exit() with the exit code
        previously defined via :py:meth:`~tcex.tcex.TcEx.exit_code` or provided
        during the call of this method.

        Args:
            code (Optional [integer]): The exit code value for the app.
            msg (Optional [string]): A message to log and add to message tc output.
        """
        # add exit message to message.tc file and log
        if msg is not None:
            if code in [0, 3] or (code is None and self.exit_code in [0, 3]):
                self.log.info(msg)
            else:
                self.log.error(msg)
            self.message_tc(msg)

        if code is None:
            code = self.exit_code
        elif code in [0, 1, 3]:
            pass
        else:
            self.log.error(u'Invalid exit code')
            code = 1

        if self._default_args.tc_aot_enabled:
            # push exit message
            self.playbook.aot_rpush(code)

        self.log.info(u'Exit Code: {}'.format(code))
        sys.exit(code)

    @property
    def exit_code(self):
        """Return the current exit code."""
        return self._exit_code

    @exit_code.setter
    def exit_code(self, code):
        """Set the App exit code.

        For TC Exchange Apps there are 3 supported exit codes.
        * 0 indicates a normal exit
        * 1 indicates a failure during execution
        * 3 indicates a partial failure

        Args:
            code (integer): The exit code value for the app.
        """
        if code is not None and code in [0, 1, 3]:
            self._exit_code = code
        else:
            self.log.warning(u'Invalid exit code')

    @property
    def group_types(self):
        """Return all defined ThreatConnect Group types.

        Returns:
            (list): A list of ThreatConnect Group types.
        """
        return [
            'Adversary',
            'Campaign',
            'Document',
            'Email',
            'Event',
            'Incident',
            'Intrusion Set',
            'Signature',
            'Report',
            'Threat',
            'Task'
        ]

    def handle_error(self, code, message_values=None, raise_error=True):
        """Raise RuntimeError

        Args:
            code (integer): The error code from API or SDK.
            message (string): The error message from API or SDK.
        """
        try:
            if message_values is None:
                message_values = []
            message = self.error_codes.message(code).format(*message_values)
            self.log.error('Error code: {}, {}'.format(code, message))
        except AttributeError:
            self.log.error('Incorrect error code provided ({}).'.format(code))
            raise RuntimeError(1000, 'Generic Failure, see logs for more details.')
        except IndexError:
            self.log.error('Incorrect message values provided for error code {} ({}).'.format(
                code, message_values))
            raise RuntimeError(1000, 'Generic Failure, see logs for more details.')
        if raise_error:
            raise RuntimeError(code, message)

    @property
    def indicator_associations_types_data(self):
        """Return ThreatConnect associations type data.

        Retrieve the data from the API if it hasn't already been retrieved.

        Returns:
            (dictionary): A dictionary of ThreatConnect associations types.
        """
        if not self._indicator_associations_types_data:
            self._association_types()  # load custom indicator associations
        return self._indicator_associations_types_data

    @property
    def indicator_types(self):
        """Return ThreatConnect Indicator types.

        Retrieve the data from the API if it hasn't already been retrieved.

        Returns:
            (list): A list of ThreatConnect Indicator types.
        """
        if not self._indicator_types:
            self._resources(True)  # load custom indicator associations
        return self._indicator_types

    @property
    def indicator_types_data(self):
        """Return ThreatConnect indicator types data.

        Retrieve the data from the API if it hasn't already been retrieved.

        Returns:
            (dict): A dictionary of ThreatConnect Indicator data.
        """
        if not self._indicator_types_data:
            # load custom indicator associations
            self._resources(True)
        return self._indicator_types_data

    def inject_params(self, params):
        """Inject params into sys.argv from secureParams API, AOT, or user provided."""
        default_bool_args = [
            'apply_proxy_external',
            'apply_proxy_ext',
            'apply_proxy_tc',
            'batch_halt_on_error',
            'tc_aot_enabled',
            'tc_log_to_api',
            'tc_proxy_external',
            'tc_proxy_tc',
            'tc_secure_params',
            'tc_verify'
        ]
        for arg, value in params.items():
            cli_arg = '--{}'.format(arg)
            if cli_arg in sys.argv:
                # arg already passed on the command line
                self.log.debug('skipping existing arg: {}'.format(cli_arg))
                continue

            # ThreatConnect secure/AOT params should be updated in the future to proper JSON format.
            # MultiChoice data should be represented as JSON array and Boolean values should be a
            # JSON boolean and not a string.
            param_data = self.install_json_params.get(arg, {})
            if param_data.get('type', '').lower() == 'multichoice':
                # update "|" delimited value to a proper array for params that have type of
                # MultiChoice.
                value = value.split('|')
            elif param_data.get('type', '').lower() == 'boolean':
                # update value to be a boolean instead of string "true"/"false".
                value = self.utils.to_bool(value)
            elif arg in default_bool_args:
                value = self.utils.to_bool(value)

            if isinstance(value, (bool)):
                # handle bool values as flags (e.g., --flag) with no value
                if value:
                    sys.argv.append(cli_arg)
            elif isinstance(value, (list)):
                for mcv in value:
                    sys.argv.append(cli_arg)
                    sys.argv.append('{}'.format(mcv))
            else:
                sys.argv.append(cli_arg)
                sys.argv.append('{}'.format(value))

        # reset default_args now that values have been injected into sys.argv
        self._default_args, unknown = self.parser.parse_known_args()

        # reinitialize logger with new log level and api settings
        self._logger()

    @property
    def install_json(self):
        """Return contents of install.json configuration file, loading from disk if required."""
        if self._install_json is None:
            try:
                install_json_filename = os.path.join(os.getcwd(), 'install.json')
                with open(install_json_filename, 'r') as fh:
                    self._install_json = json.load(fh)
            except IOError:
                self.log.warning(u'Could not retrieve App Data.')
        return self._install_json

    @property
    def install_json_params(self):
        """Parse params from install.json into a dict by name."""
        if not self._install_json_params:
            for param in self.install_json.get('params', []):
                self._install_json_params[param.get('name')] = param
        return self._install_json_params

    def job(self):
        """**[Deprecated]** Return instance of Job module

        .. warning:: The job module is deprecated and will be removed in TcEx version 0.9.0. Use
                     tcex.batch instead.
        """
        self.log.warning('Jobs module will be deprecated in TcEx version 0.9.0.')
        from .tcex_job import TcExJob
        return TcExJob(self)

    @property
    def jobs(self):
        """**[Deprecated]** Include the jobs Module.

        .. warning:: The job module is deprecated and will be removed in TcEx version 0.9.0. Use
                     tcex.batch instead.
        """
        if self._jobs is None:
            from .tcex_job import TcExJob
            self._jobs = TcExJob(self)
        return self._jobs

    def metric(self, name, description, data_type, interval, keyed=False):
        """Get instance of the Metrics module.

        Args:
            name (string): The name for the metric.
            description (string): The description of the metric.
            data_type (string): The type of metric: Sum, Count, Min, Max, First, Last, and Average.
            interval (string): The metric interval: Hourly, Daily, Weekly, Monthly, and Yearly.
            keyed (boolean): Indicates whether the data will have a keyed value.

        Returns:
            (object): An instance of the Metrics Class.
        """
        from .tcex_metrics_v2 import TcExMetricsV2
        return TcExMetricsV2(self, name, description, data_type, interval, keyed)

    def notification(self):
        """Get instance of the Notification module.

        Returns:
            (object): An instance of the Notification Class.
        """
        from .tcex_notification_v2 import TcExNotificationV2
        return TcExNotificationV2(self)

    def message_tc(self, message, max_length=255):
        """Write data to message_tc file in TcEX specified directory.

        This method is used to set and exit message in the ThreatConnect Platform.
        ThreatConnect only supports files of max_message_length.  Any data exceeding
        this limit will be truncated by this method.

        Args:
            message (string): The message to add to message_tc file
        """
        if os.access(self.default_args.tc_out_path, os.W_OK):
            message_file = '{}/message.tc'.format(self.default_args.tc_out_path)
        else:
            message_file = 'message.tc'

        message = '{}\n'.format(message)
        if max_length - len(message) > 0:
            with open(message_file, 'a') as mh:
                mh.write(message)
        elif max_length > 0:
            with open(message_file, 'a') as mh:
                mh.write(message[:max_length])
        max_length -= len(message)

    @property
    def playbook(self):
        """Include the Playbook Module.

        .. Note:: Playbook methods can be accessed using ``tcex.playbook.<method>``.
        """
        if self._playbook is None:
            from .tcex_playbook import TcExPlaybook
            self._playbook = TcExPlaybook(self)
        return self._playbook

    @property
    def proxies(self):
        """Formats proxy configuration into required format for Python Requests module.

        Generates a dictionary for use with the Python Requests module format
        when proxy is required for remote connections.

        **Example Response**
        ::

            {"http": "http://user:pass@10.10.1.10:3128/"}

        Returns:
           (dictionary): Dictionary of proxy settings
        """
        proxies = {}
        if (self.default_args.tc_proxy_host is not None and
                self.default_args.tc_proxy_port is not None):

            if (self.default_args.tc_proxy_username is not None and
                    self.default_args.tc_proxy_password is not None):
                tc_proxy_username = quote(self.default_args.tc_proxy_username, safe='~')
                tc_proxy_password = quote(self.default_args.tc_proxy_password, safe='~')

                proxies = {
                    'http': 'http://{}:{}@{}:{}'.format(
                        tc_proxy_username, tc_proxy_password,
                        self.default_args.tc_proxy_host, self.default_args.tc_proxy_port),
                    'https': 'http://{}:{}@{}:{}'.format(
                        tc_proxy_username, tc_proxy_password,
                        self.default_args.tc_proxy_host, self.default_args.tc_proxy_port)
                }
            else:
                proxies = {
                    'http': 'http://{}:{}'.format(
                        self.default_args.tc_proxy_host, self.default_args.tc_proxy_port),
                    'https': 'http://{}:{}'.format(
                        self.default_args.tc_proxy_host, self.default_args.tc_proxy_port)
                }
        return proxies

    def request(self, session=None):
        """Return an instance of the Request Class.

        A wrapper on the Python Request Module specifically for interacting with the
        ThreatConnect API.  However, this can also be used for connecting to other
        API endpoints.

        Returns:
            (object): An instance of Request Class
        """
        try:
            from .tcex_request import TcExRequest
            return TcExRequest(self, session)
        except ImportError as e:
            self.handle_error(105, [e])

    def request_external(self):
        """Return an instance of the Request Class with Proxy Set

        See :py:mod:`~tcex.tcex.TcEx.request`

        Returns:
            (object): An instance of Request Class
        """
        r = self.request()
        if self.default_args.tc_proxy_external:
            self.log.info(u'Using proxy server for external request {}:{}.'.format(
                self.default_args.tc_proxy_host, self.default_args.tc_proxy_port))
            r.proxies = self.proxies
        return r

    def request_tc(self):
        """**[Deprecated]** Return an instance of the Request Class with Proxy and Authorization Set

        .. warning:: This method is deprecated and will be removed in TcEx version 0.9.0. Use
                     tcex.session instead.

        See :py:mod:`~tcex.tcex.TcEx.request`

        Returns:
            (object): An instance of Request Class
        """
        self.log.warning('This App is using a deprecated method (request_tc).')
        r = self.request(self.session)
        return r

    def resource(self, resource_type):
        """Get instance of Resource Class with dynamic type.

        Args:
            resource_type: The resource type name (e.g Adversary, User Agent, etc).

        Returns:
            (object): Instance of Resource Object child class.
        """
        try:
            resource = getattr(self.resources, self.safe_rt(resource_type))(self)
        except AttributeError:
            self._resources(True)
            resource = getattr(self.resources, self.safe_rt(resource_type))(self)
        return resource

    def results_tc(self, key, value):
        """Write data to results_tc file in TcEX specified directory.

        The TcEx platform support persistent values between executions of the App.  This
        method will store the values for TC to read and put into the Database.

        Args:
            key (string): The data key to be stored.
            value (string): The data value to be stored.
        """
        if os.access(self.default_args.tc_out_path, os.W_OK):
            results_file = '{}/results.tc'.format(self.default_args.tc_out_path)
        else:
            results_file = 'results.tc'

        new = True
        open(results_file, 'a').close()  # ensure file exists
        with open(results_file, 'r+') as fh:
            results = ''
            for line in fh.read().strip().split('\n'):
                if not line:
                    continue
                try:
                    k, v = line.split(' = ')
                except ValueError:
                    # handle null/empty value (e.g., "name =")
                    k, v = line.split(' =')
                if k == key:
                    v = value
                    new = False
                if v is not None:
                    results += '{} = {}\n'.format(k, v)
            if new and value is not None:  # indicates the key/value pair didn't already exist
                results += '{} = {}\n'.format(key, value)
            fh.seek(0)
            fh.write(results)
            fh.truncate()

    def results_tc_args(self):
        """Read data from results_tc file from previous run of app.

        This method is only required when not running from the with the
        TcEX platform and is only intended for testing apps locally.

        Returns:
            (dictionary): A dictionary of values written to results_tc.
        """
        results = []
        if os.access(self.default_args.tc_out_path, os.W_OK):
            result_file = '{}/results.tc'.format(self.default_args.tc_out_path)
        else:
            result_file = 'results.tc'
        if os.path.isfile(result_file):
            with open(result_file, 'r') as rh:
                results = rh.read().strip().split('\n')
            os.remove(result_file)
        for line in results:
            if not line or ' = ' not in line:
                continue
            key, value = line.split(' = ')
            if value == 'true':
                value = True
            elif value == 'false':
                value = False
            elif not value:
                value = None
            setattr(self.default_args, key, value)

    def s(self, data, errors='strict'):
        """Decode value using correct Python 2/3 method.

        This method is intended to replace the :py:meth:`~tcex.tcex.TcEx.to_string` method with
        better logic to handle poorly encoded unicode data in Python2 and still work in Python3.

        Args:
            data (any): Data to ve validated and (de)encoded
            errors (string): What method to use when dealing with errors.

        Returns:
            (string): Return decoded data
        """
        try:
            if data is None or isinstance(data, (int, list, dict)):
                pass  # Do nothing with these types
            elif isinstance(data, unicode):
                try:
                    data.decode('utf-8')
                except UnicodeEncodeError:  # 2to3 converts unicode to str
                    # 2to3 converts unicode to str
                    data = str(data.encode('utf-8').strip(), errors=errors)
                    self.log.warning(u'Encoding poorly encoded string ({})'.format(data))
                except AttributeError:
                    pass  # Python 3 can't decode a str
            else:
                data = str(data, 'utf-8', errors=errors)  # 2to3 converts unicode to str
        except NameError:
            pass  # Can't decode str in Python 3
        return data

    def safe_indicator(self, indicator, errors='strict'):
        """Indicator encode value for safe HTTP request.

        Args:
            indicator (string): Indicator to URL Encode
            errors (string): The error handler type.

        Returns:
            (string): The urlencoded string
        """
        if indicator is not None:
            try:
                indicator = quote(self.s(str(indicator), errors=errors), safe='~')
            except KeyError:
                indicator = quote(bytes(indicator), safe='~')
        return indicator

    def epoch_seconds(self, delta=None):
        """**[Deprecated]** Get epoch seconds for now or using a time delta.

        .. warning:: This method is deprecated and will be removed in TcEx version 0.9.0.  Use the
                     tcex.utils date methods instead.

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1

            {'days': 1}
            {'weeks': 3}
            {'months': 4}
            {'days': 1, 'weeks': 3, 'months': 4}


        .. Note:: More information can be found at
                  https://dateutil.readthedocs.io/en/stable/relativedelta.html

        Args:
            delta (Optional [integer]): The exit code value for the app.
        Returns:
            (int): A integer representing epoch seconds.
        """
        from datetime import datetime
        from dateutil.relativedelta import relativedelta
        self.log.warning('This App is using a deprecated method (epoch seconds).')
        epoch = datetime.now()
        if delta is not None:
            epoch = epoch - relativedelta(**delta)

        return int(time.mktime(epoch.timetuple()))

    @staticmethod
    def expand_indicators(indicator):
        """Process indicators expanding file hashes/custom indicators into multiple entries.

        Args:
            indicator (string): " : " delimited string
        Returns:
            (list): a list of indicators split on " : ".
        """
        if indicator.count(' : ') > 0:
            # handle all multi-valued indicators types (file hashes and custom indicators)
            indicator_list = []

            # group 1 - lazy capture everything to first <space>:<space> or end of line
            iregx_pattern = r'^(.*?(?=\s\:\s|$))?'
            iregx_pattern += r'(?:\s\:\s)?'  # remove <space>:<space>
            # group 2 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead (optional <space>):<space> or end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=(?:\s)?\:\s|$))?'
            iregx_pattern += r'(?:(?:\s)?\:\s)?'  # remove (optional <space>):<space>
            # group 3 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=$))?$'
            iregx = re.compile(iregx_pattern)

            indicators = iregx.search(indicator)
            if indicators is not None:
                indicator_list = list(indicators.groups())
        else:
            # handle all single valued indicator types (address, host, etc)
            indicator_list = [indicator]

        return indicator_list

    @staticmethod
    def safe_rt(resource_type, lower=False):
        """Format the Resource Type.

        Takes Custom Indicator types with a space character and return a *safe* string.

        (e.g. *User Agent* is converted to User_Agent or user_agent.)

        Args:
           resource_type (string): The resource type to format.
           lower (boolean): Return type in all lower case

        Returns:
            (string): The formatted resource type.
        """
        if resource_type is not None:
            resource_type = resource_type.replace(' ', '_')
            if lower:
                resource_type = resource_type.lower()
        return resource_type

    @staticmethod
    def safe_group_name(group_name, group_max_length=100, ellipsis=True):
        """Truncate group name to match limit breaking on space and optionally add an ellipsis.

        .. note:: Currently the ThreatConnect group name limit is 100 characters.

        Args:
           group_name (string): The raw group name to be truncated.
           group_max_length (int): The max length of the group name.
           ellipsis (boolean): If true the truncated name will have '...' appended.

        Returns:
            (string): The truncated group name with optional ellipsis.
        """
        ellipsis_value = ''
        if ellipsis:
            ellipsis_value = ' ...'

        if group_name is not None and len(group_name) > group_max_length:
            # split name by spaces and reset group_name
            group_name_array = group_name.split(' ')
            group_name = ''
            for word in group_name_array:
                word = u'{}'.format(word)
                if (len(group_name) + len(word) + len(ellipsis_value)) >= group_max_length:
                    group_name = '{}{}'.format(group_name, ellipsis_value)
                    group_name = group_name.lstrip(' ')
                    break
                group_name += ' {}'.format(word)
        return group_name

    def safetag(self, tag, errors='strict'):
        """URL Encode and truncate tag to match limit (128 characters) of ThreatConnect API.

        Args:
           tag (string): The tag to be truncated

        Returns:
            (string): The truncated tag
        """
        if tag is not None:
            try:
                # handle unicode characters and url encode tag value
                tag = quote(self.s(tag[:128], errors=errors), safe='~')
            except KeyError as e:
                warn = 'Failed converting tag to safetag ({})'.format(e)
                self.log.warning(warn)
        return tag

    def safeurl(self, url, errors='strict'):
        """URL encode value for safe HTTP request.

        Args:
            url (string): The string to URL Encode.

        Returns:
            (string): The urlencoded string.
        """
        if url is not None:
            url = quote(self.s(url, errors=errors), safe='~')
        return url

    @property
    def session(self):
        """Return an instance of Requests Session configured for the ThreatConnect API."""
        if self._session is None:
            from .tcex_session import TcExSession
            self._session = TcExSession(self)
        return self._session

    def to_string(self, data, errors='strict'):
        """**[Deprecated]** Decode value using correct Python 2/3 method

        .. warning:: This method is deprecated and will be removed in TcEx version 0.9.0.

        Args:
            data (any): Data to ve validated and re-encoded
            errors (string): What method to use when dealing with errors.

        Returns:
            (string): Return decoded data

        """
        self.log.warning('This App is using a deprecated method (to_string).')
        if data is not None and not isinstance(data, str):
            data = str(data, 'utf-8', errors=errors)
        return data
